/*
 * Loesung zu Blatt 12
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

/* ein bisschen groesser als die maximale Laenge einer Zeile ... */
#define MAXLEN 1024

/*
 * Zerlegt einen String s in (mgl. leere!)
 * Tokens, die voneinander durch je ein delim-Zeichen
 * getrennt sind.
 *
 * Unterschied zu strtok(): Es sind bei strtok() mehrere
 * Trenner mgl.; ausserdem gibt's bei strtok() _keine_ leeren
 * Token
 */
char *mytok(char *s, char delim) {
    static char *old = NULL;
    char *buf = s ? s : old;	/* im Fall von s = NULL verwenden wir old! */
    if (buf == NULL)		/* am Ende? => return NULL */
        return NULL;
    if ((old = strchr(buf, delim)) != NULL) { /* noch mehr als ein Token? */
        *old = '\0';		/* Nullbyte zwischen akt. und folgendem Token einfuegen */
        old++;			/* old zeigt jetzt schon auf das folgende Token! */
    }
    return buf;
}

/*
 * fuehrt die Spaltenextraktion durch
 * liest von stdin
 * und schreibt nach stdout
 *
 * delim enthaelt das Trennzeichen
 * und fields ist ein Array von Wahrheitswerten,
 * wobei fields[i] angibt, ob die i-te (i >= 0!)
 * Spalte ausgegeben werden soll.
 */
void cut(char delim, int *fields) {
    char line[MAXLEN];

    /* Zeile fuer Zeile von stdin lesen */
    while (fgets(line, MAXLEN, stdin) != NULL) {
        int j = 0, anz = 0;
        char *token;
        int len = strlen(line);	/* Laenge der Zeile berechnen */

        if (line[len-1] == '\n')	/* Newline am Ende entfernen */
            line[len-1] = '\0';

	/*
	 * Zeile so wie bei strtok() zerlegen in die
	 * einzelnen Spalten
	 */
        token = mytok(line, delim);
        while (token) {
	    /* soll diese Spalte ausgegeben werden? */
            if (fields[j++]) {
                if (anz++ > 0)			/* gab's vorher schon eine Spalte in der Zeile? */
                    fputc(delim, stdout);	/* => dann Spaltentrenner ausgeben */
                fputs(token, stdout);		/* Attributwert ausgeben */
            }
            token = mytok(NULL, delim);		/* ... und zum naechsten Token */
        }
	
        puts("");			/* zum Schluss dann noch ein Newline ans Ende der Zeile */
    }
}

int main(int argc, char **argv) {
    char delim = '\t';		/* Trennzeichen; default = Tabulator */
    int fields[MAXLEN];		/* Array fuer die auszugebenden Spalten; Indizierung ab 0! */
    int optf = 0;		/* gab's die Option -f ? */
    int errflag = 0;		/* gab's einen Fehler bei der Optionsverarbeitung? */
    int opt;
    int i;

    /* zunaechst mal das Array fields initialisieren => keine Spalte wird ausgegeben
     * fuer jede auszugebende Spalte muss eine 1 eingetragen werden */
    for (i = 0; i < MAXLEN; i++)
        fields[i] = 0;

    /*
     * Optionen verarbeiten (-d, -f und -o jeweils mit Parameter)
     */
    while ((opt = getopt(argc, argv, "d:f:o:")) != EOF) {
        switch (opt) {
            case 'd' :
		/* besteht das Argument zu dieser Option nur aus einem Zeichen? */
                if (optarg[0] != '\0' && optarg[1] == '\0')
                    delim = optarg[0];
                else {
                    fprintf(stderr, "%s: the delimiter must be a single character\n", argv[0]);
                    return 1;
                }
                break;
            case 'f' :
                {
		    /*
		     * Argument zu dieser Option mittels mytok()
		     * so wie mit strtok() zerlegen
		     */
                    char *token;
                    token = mytok(optarg, ',');
                    while (token) {
                        int num, n;
			/* jedes Token sollte eine Integer >= 1 sein,
			 * dies testen wir mit sscanf() und %n
			 * (siehe Beispielprogramm)
			 */
                        if (sscanf(token, "%d%n", &num, &n) != 1 || token[n] != '\0' || num < 1) {
                            fprintf(stderr, "%s: %s: wrong column index format\n", argv[0], token);
                            errflag = 1;
                        }
                        else if (num <= MAXLEN)
                            fields[num-1] = 1;	/* falls Integer korrekt => Eintrag in fields[] */
                        token = mytok(NULL, ',');	/* ... und zum naechsten Token */
                    }
                }
                optf = 1;	/* merken, dass Option -f dran war! */
                break;
            case 'o' :
		/* Standardausgabe in die angegebene Datei umlenken */
                close(1);
                if (open(optarg, O_WRONLY | O_CREAT | O_TRUNC, 0666) < 0) {
		    char err[1024];
		    sprintf(err, "%s: %s", argv[0], optarg);
            	    perror(err);
                    exit(2);
		}
                break;
            case '?' :
		/* Fehler bei der Optionsverarbeitung */
                errflag = 1;
                break;
        }
    }

    /* Fehlte die Option -f => Fehlermeldung */
    if (!optf)
        fprintf(stderr, "%s: list of fields is missing\n", argv[0]);

    /* Fehler bei der Optionsverarbeitung oder -f fehlt? => Usage-Meldung */
    if (errflag || !optf) {
        fprintf(stderr, "Usage: %s [-d <delim>] [-f <field list>] [-o <file>] {<infile>}\n", argv[0]);
        return 3;
    }

    /* gibts noch weitere Kommandozeilen-Argumente? */
    if (optind < argc)
	/* ja => diese nacheinander abarbeiten */
        for (i = optind; i < argc; i++) {
	    /* Standardeingabe aus der angegebenen Datei beziehen ... (Umlenkung!) */
            close(0);
            if (open(argv[i], O_RDONLY) < 0) {
		char err[1024];
		sprintf(err, "%s: %s", argv[0], argv[i]);
                perror(err);
	    }

	    /* und cut durchfuehren */
            cut(delim, fields);
        }
    else
	/* nein => cut durchfuehren, ohne stdin umzulenken */
        cut(delim, fields);

    return 0;
}
